// AST

use nom_locate::LocatedSpan;

pub type Span<'a> = LocatedSpan<&'a str>;

#[derive(Debug, Clone, Copy, PartialEq)]
pub enum Op {
    Eq,
    Neq,
    And,
    Or,
    Add,
    Sub,
    Mul,
    Div,
    Pow,
    Not,
}

type SpanOp<'a> = (Span<'a>, Op);

pub type SpanId<'a> = Span<'a>;

#[derive(Debug, Clone, PartialEq)]
pub enum Expr<'a> {
    Num(i32),
    Bool(bool),
    Par(Box<SpanExpr<'a>>),
    Id(String),
    Call(String, Vec<SpanExpr<'a>>),
    BinOp(Op, Box<SpanExpr<'a>>, Box<SpanExpr<'a>>),
    UnaryOp(Op, Box<SpanExpr<'a>>),
}

pub type SpanExpr<'a> = (Span<'a>, Expr<'a>);

#[derive(Debug, PartialEq, Clone)]
pub enum Cmd<'a> {
    // let <mut> id : <& <mut>>Type = expr
    Let(Mutability, SpanId<'a>, Type, SpanExpr<'a>),
    // id = expr
    Assign(SpanExpr<'a>, SpanExpr<'a>),
    // if predicate do-this, and optionally do-that)
    If(SpanExpr<'a>, Block<'a>, Option<Block<'a>>),
    // while predicate do-this
    While(SpanExpr<'a>, Block<'a>),
    Return(SpanExpr<'a>),
}

pub type SpanCmd<'a> = (Span<'a>, Cmd<'a>);

#[derive(Debug, PartialEq, Clone)]
pub enum Mutability {
    Imm,
    Mut,
}
pub type SpanMut<'a> = (Span<'a>, Mutability);

pub type SpanBlock<'a> = (Span<'a>, Vec<SpanCmd<'a>>);
pub type Block<'a> = Vec<Cmd<'a>>;

#[derive(Debug, PartialEq, Clone)]
pub enum Type {
    I32,
    Bool,
    Unit,
    Mut(Box<Type>),
    Ref(Box<Type>),
    // no structs
}

// pub type SpanType<'a> = (Span<'a>, Type<'a>);

#[derive(Debug, PartialEq, Clone)]
pub struct Func<'a> {
    pub sig: (SpanId<'a>, Vec<(SpanId<'a>, Type)>, Type),
    pub body: Block<'a>,
}

#[derive(Debug, PartialEq, Clone)]
pub enum Item<'a> {
    Func(Func<'a>),
}

pub type Prog<'a> = Vec<Item<'a>>;
